import os
import re
import psycopg2
from psycopg2 import OperationalError
from psycopg2.extensions import connection
from typing import Tuple, List, Optional, Union

from lineagex.utils import produce_json
from lineagex.SqlToDict import SqlToDict
from lineagex.stack import *
from lineagex.ColumnLineage import ColumnLineage


class LineageXWithConn:
    def __init__(
        self,
        path: Optional[Union[List, str]] = None,
        search_schema: Optional[str] = "public",
        url: Optional[str] = "",
        username: Optional[str] = "",
        password: Optional[str] = "",
    ) -> None:
        self.part_tables = None
        self.df = None
        self.conn_string = (
            url.split("//")[0]
            + "//"
            + username
            + ":"
            + password
            + "@"
            + url.split("//")[1]
        )
        self.schema = search_schema.split(",")[0]
        self.search_schema = search_schema
        self.new_view_list = []
        self.s = Stack()
        self.path = path
        self.sql_files_dict = {}
        self.creation_list = []
        self.finished_list = []
        self.output_dict = {}
        self._run_table_lineage()

    def _run_table_lineage(self) -> None:
        """
        The driver function to extract the table lineage information
        :return: output an interactive html for the table lineage information
        """
        self.conn = self._check_db_connection()
        self.conn.autocommit = True
        self.part_tables, self.schema_list = self._get_part_tables()
        self.sql_files_dict = SqlToDict(self.path, self.schema_list).sql_files_dict
        for name, sql in self.sql_files_dict.items():
            try:
                if name not in self.finished_list:
                    self._explain_sql(name=name, sql=sql)
                else:
                    continue
            except Exception as e:
                print("{} is not processed because it countered {}".format(name, e))
                continue
        produce_json(output_dict=self.output_dict, engine=self.conn, search_schema=self.search_schema)
        self._delete_view()
        self.conn.close()

    def _delete_view(self) -> None:
        """
        Delete all temporary tables in the new_view_list
        :return: None
        """
        # reverse it just in case to drop dependencies first
        self.new_view_list = self.new_view_list[::-1]
        cur = self.conn.cursor()
        for i in self.new_view_list:
            cur.execute("""DROP TABLE {} CASCADE""".format(i))
            print(i + " dropped")
        cur.close()

    def _create_view(self, name: Optional[str] = "", sql: Optional[str] = "") -> None:
        """
        Create temporary tables with no data from the given sql
        :param name: name of the table
        :param sql: the sql for the table
        :param conn: the psycopg2 connection it was using
        :return: None
        """
        # connect and create view
        cur = self.conn.cursor()
        cur.execute(
            """SET search_path TO {}, {};""".format(self.search_schema, self.schema)
        )
        if sql.endswith(";"):
            cur.execute(
                """CREATE TABLE {}.{} AS {} WITH NO DATA;""".format(
                    self.schema, name, sql[:-1]
                )
            )
        else:
            cur.execute(
                """CREATE TABLE {}.{} AS {} WITH NO DATA;""".format(
                    self.schema, name, sql
                )
            )
        cur.close()
        print(self.schema + "." + name + " created")

    def _explain_sql(self, name: Optional[str] = "", sql: Optional[str] = "") -> None:
        """
        Main function for extracting the table name from the sql. It tries to explain the current file's sql by
        analyzing the logical plan. But if its dependency is missing, the current one is put onto a stack and checking
        on the dependency table, then pops the stack when the current one is done.
        :param name: name of the file
        :param sql: the sql from the file
        :return: updates file_list, sql_list, table_list, new_view_list
        """
        try:
            print(name)
            cur = self.conn.cursor()
            cur.execute("""SET search_path TO {};""".format(self.search_schema))
            cur.execute(
                """EXPLAIN (VERBOSE TRUE, FORMAT JSON, COSTS FALSE) {}""".format(sql)
            )
            log_plan = cur.fetchall()
            cur.close()
            while True:
                if isinstance(log_plan, list) or isinstance(log_plan, tuple):
                    log_plan = log_plan[0]
                elif isinstance(log_plan, dict):
                    log_plan = log_plan["Plan"]
                    break
            if name in self.creation_list:
                self._create_view(name=name, sql=sql)
                self.new_view_list.append(self.schema + "." + name)
                self.creation_list.pop(self.creation_list.index(name))
                table_name = self.schema + "." + name
            else:
                cur = self.conn.cursor()
                cur.execute("""SET search_path TO {};""".format(self.search_schema))
                cur.execute(
                    """SELECT CONCAT (schemaname,'.', tablename) from pg_tables WHERE schemaname = ANY('{{{0}}}') and tablename = '{1}'""".format(
                        self.search_schema, name
                    )
                )
                table_name = cur.fetchone()
                cur.close()
                if table_name:
                    table_name = table_name[0]
                else:
                    table_name = self.schema + "." + name
            col_lineage = ColumnLineage(
                plan=log_plan,
                sql=sql,
                table_name=table_name,
                conn=self.conn,
                part_tables=self.part_tables,
                search_schema=self.search_schema,
            )
            self.output_dict[table_name] = {
                "tables": col_lineage.table_list,
                "columns": col_lineage.column_dict,
                "table_name": table_name,
            }
            self.finished_list.append(name)
            while not self.s.isEmpty():
                if self.s.peek() in self.sql_files_dict.keys():
                    sql = self.sql_files_dict[self.s.peek()]
                    return self._explain_sql(name=self.s.pop(), sql=sql)
                else:
                    print("no sql file for the dependencies")
            return
        except psycopg2.ProgrammingError as e:
            # does not exist error code
            if e.pgcode == "42P01":
                error_msg = e.pgerror
                no_find_idx = error_msg.find("does not exist")
                relation_idx = error_msg.find("relation")
                schema_table = error_msg[relation_idx:no_find_idx]
                table_name = schema_table.split(" ")[-2].split(".")[-1].strip('\"')
                self.s.push(name)
                #print(table_name, self.sql_files_dict)
                if table_name in self.sql_files_dict.keys():
                    if self.schema + "." + table_name in self.new_view_list:
                        print(
                            "{}.{} is already created, but the created schema is different from the queried schema for {} in {}.sql".format(
                                self.schema, table_name, table_name, name
                            )
                        )
                        return
                    print(
                        name
                        + " is dependant on "
                        + table_name
                        + ", creating that first\n"
                    )
                    self.creation_list.append(table_name)
                    sql = self.sql_files_dict[table_name]
                    return self._explain_sql(name=table_name, sql=sql)
                else:
                    print(
                        name
                        + " is skipped because it is missing dependency table "
                        + table_name
                    )
                    return
            else:
                print(e)
        except Exception as e:
            print(e)

    def _check_db_connection(self) -> connection:
        """
        Check if the conn_string is good
        :return: the sqlalchemy engine
        """
        try:
            psycopg2.connect(self.conn_string)
            print("database connected")
        except OperationalError:
            print("authentication error")
        return psycopg2.connect(self.conn_string)

    def _get_part_tables(self) -> Tuple[dict, List]:
        """
        Find out the partition table in the Postgres database
        :return: the result with the parent and partitioned table
        """
        cur = self.conn.cursor()
        cur.execute(
            """SELECT
                concat_ws('.', nmsp_parent.nspname, parent.relname) AS parent,
                concat_ws('.', nmsp_child.nspname, child.relname) AS child
            FROM pg_inherits
                JOIN pg_class parent            ON pg_inherits.inhparent = parent.oid
                JOIN pg_class child             ON pg_inherits.inhrelid   = child.oid
                JOIN pg_namespace nmsp_parent   ON nmsp_parent.oid  = parent.relnamespace
                JOIN pg_namespace nmsp_child    ON nmsp_child.oid   = child.relnamespace"""
        )
        ret = cur.fetchall()
        parent_dict = {}
        for i in ret:
            parent_dict[i[1]] = i[0]
        # get the schema list as well
        cur.execute("select nspname from pg_catalog.pg_namespace;")
        result = cur.fetchall()
        schema_list = [s[0] for s in result]
        cur.close()
        return parent_dict, schema_list


if __name__ == "__main__":
    pass



from sqlglot import parse_one, exp
from sqlglot import expressions
from typing import Optional, List, Tuple
import os

shared_conditions = [exp.Where, exp.EQ, exp.Group, exp.Having, exp.Order]
shared_conditions_with_table = [exp.Where, exp.EQ, exp.Group, exp.Having, exp.Order, exp.From, exp.Join]


class ColumnLineageNoConn:
    def __init__(self, sql: Optional[str] = "", input_table_dict: Optional[dict] = None):
        self.column_dict = {}
        self.table_alias_dict = {}
        self.cte_table_dict = {}
        self.cte_dict = {}
        self.input_table_dict = input_table_dict
        self.sql_ast = parse_one(sql, read="postgres")
        self.all_used_col = []
        self.table_list = []
        self.all_subquery_table = []
        self.sub_tables = []
        self.sub_cols = []
        self._run_cte_lineage()
        # Everything other than CTEs, and pop the CTE tree
        for with_sql in self.sql_ast.find_all(exp.With):
            with_sql.pop()
        self._sub_shared_col_conds(sql_ast=self.sql_ast)
        self._run_lineage(self.sql_ast, False)
        #print(self.cte_dict)
        #print(self.column_dict)
        #print(self.table_list)

    def _run_lineage(self, sql_ast: expressions = None, subquery_flag: bool = False) -> None:
        """
        Run the lineage after all the cte and subquery are resolved
        :param sql_ast: the ast for the sql
        :param subquery_flag: check if it is a subquery
        """
        #print(sql_ast)
        if not subquery_flag:
            main_tables = self._resolve_table(part_ast=sql_ast)
            self.table_list = self._find_all_tables(temp_table_list=main_tables)
            self.table_list.extend(self.all_subquery_table)
            self.all_used_col = []
            self._shared_col_conds(part_ast=sql_ast, used_tables=main_tables)
            self.all_used_col.extend(self.sub_cols)
            if isinstance(sql_ast, exp.Union) or isinstance(sql_ast, exp.Except) or isinstance(sql_ast, exp.Intersect):
                for col in sql_ast.find_all(exp.Column):
                    self.all_used_col.extend(self._find_alias_col(col_sql=col.sql(), temp_table=main_tables))
            self.all_used_col = set(self.all_used_col)
            if sql_ast.find(exp.Select):
                for projection in sql_ast.find(exp.Select).expressions:
                    col_name = projection.alias_or_name
                    self.column_dict = self._resolve_proj(projection=projection, col_name=col_name, target_dict=self.column_dict, source_table=main_tables)
        else:
            temp_sub_cols = []
            for col in sql_ast.find_all(exp.Column):
                temp_sub_cols.extend(self._find_alias_col(col_sql=col.sql(), temp_table=self.sub_tables))
            self.sub_cols.extend(temp_sub_cols)
            #print(temp_sub_cols)

    def _sub_shared_col_conds(self, sql_ast: expressions = None) -> None:
        """
        After the cte are resolved, run the subquery ast that is with the shared conditions(WHERE, GROUP BY, etc)
        :param sql_ast: the ast without the cte
        """
        # add in more conditions, including FROM/JOIN
        for cond in shared_conditions_with_table:
            for cond_sql in sql_ast.find_all(cond):
                for sub_ast in cond_sql.find_all(exp.Subquery):
                    self.sub_tables = self._resolve_table(part_ast=sub_ast)
                    self.all_subquery_table.extend(self._find_all_tables(temp_table_list=self.sub_tables))
                    self._run_lineage(sub_ast, True)
                    sub_ast.pop()

    def _sub_shared_col_conds_cte(self, sql_ast: expressions = None) -> Tuple[List, List]:
        """
        Run the subquery inside the cte first that is with the shared conditions(WHERE, GROUP BY, etc)
        :param sql_ast: the ast for the cte
        """
        all_cte_sub_table = []
        all_cte_sub_cols = []
        # add in more conditions, including FROM/JOIN
        for cond in shared_conditions_with_table:
            for cond_sql in sql_ast.find_all(cond):
                for sub_ast in cond_sql.find_all(exp.Select):
                    temp_sub_table = self._resolve_table(part_ast=sub_ast)
                    temp_sub_cols = []
                    for col in sub_ast.find_all(exp.Column):
                        temp_sub_cols.extend(self._find_alias_col(col_sql=col.sql(), temp_table=temp_sub_table))
                    all_cte_sub_table.extend(self._find_all_tables(temp_table_list=temp_sub_table))
                    all_cte_sub_cols.extend(temp_sub_cols)
                    sub_ast.pop()
        return all_cte_sub_table, all_cte_sub_cols

    def _run_cte_lineage(self):
        """
        Run the lineage information for all the cte
        """
        for cte in self.sql_ast.find_all(exp.CTE):
            all_cte_sub_table, all_cte_sub_cols = self._sub_shared_col_conds_cte(sql_ast=cte)
            self.all_used_col = []
            temp_cte_dict = {}
            temp_cte_table = self._resolve_table(part_ast=cte)
            cte_name = cte.find(exp.TableAlias).alias_or_name
            self.cte_table_dict[cte_name] = list(set(self._find_all_tables(temp_table_list=temp_cte_table) + all_cte_sub_table))
            # Resolving shared conditions
            self._shared_col_conds(part_ast=cte, used_tables=temp_cte_table)
            self.all_used_col.extend(all_cte_sub_cols)
            self.all_used_col = set(self.all_used_col)
            # Resolving the projection
            for projection in cte.find(exp.Select).expressions:
                col_name = projection.alias_or_name
                temp_cte_dict = self._resolve_proj(projection=projection, col_name=col_name, target_dict=temp_cte_dict, source_table=temp_cte_table)
            self.cte_dict[cte_name] = temp_cte_dict

    def _resolve_proj(self, projection: expressions = None, col_name: Optional[str] = "", target_dict: Optional[dict] = None, source_table: Optional[List] = None) -> dict:
        """
        Resolve the projection given the projection expression
        :param projection: the given projection
        :param col_name: the column name
        :param target_dict: the dict it is outputting to
        :param source_table: all the source tables that this column might originate from
        """
        # Resolve count(*) with no alias, potentially other aggregations, MIN, MAX, SUM
        if isinstance(projection, exp.Count) or isinstance(projection.unalias(), exp.Count)\
                or isinstance(projection, exp.Avg) or isinstance(projection.unalias(), exp.Avg)\
                or isinstance(projection, exp.Max) or isinstance(projection.unalias(), exp.Max)\
                or isinstance(projection, exp.Min) or isinstance(projection.unalias(), exp.Min)\
                or isinstance(projection, exp.Sum) or isinstance(projection.unalias(), exp.Sum):
            if isinstance(projection, exp.Count):
                col_name = "count"
                self._resolve_agg_star(col_name=col_name, projection=projection, used_tables=source_table)
            elif isinstance(projection, exp.Avg):
                col_name = "avg"
                self._resolve_agg_star(col_name=col_name, projection=projection, used_tables=source_table)
            elif isinstance(projection, exp.Max):
                col_name = "max"
                self._resolve_agg_star(col_name=col_name, projection=projection, used_tables=source_table)
            elif isinstance(projection, exp.Min):
                col_name = "min"
                self._resolve_agg_star(col_name=col_name, projection=projection, used_tables=source_table)
            elif isinstance(projection, exp.Sum):
                col_name = "sum"
                self._resolve_agg_star(col_name=col_name, projection=projection, used_tables=source_table)
            else:
                self._resolve_agg_star(col_name=col_name, projection=projection.unalias(), used_tables=source_table)
        else:
            proj_columns = []
            # Resolve only *
            if not isinstance(projection, exp.Column) and projection.find(exp.Star):
                for t_name in source_table:
                    if t_name in self.input_table_dict.keys():
                        star_cols = self.input_table_dict[t_name]
                        # every column from there will be a column with that name
                        for per_star_col in star_cols:
                            target_dict[per_star_col] = sorted(
                                list(set(self._find_alias_col(col_sql=per_star_col, temp_table=source_table)).union(
                                    self.all_used_col)))
                    elif t_name in self.cte_dict.keys():
                        star_cols = list(self.cte_dict[t_name].keys())
                        for per_star_col in star_cols:
                            target_dict[per_star_col] = sorted(
                                list(set(self.cte_dict[t_name][per_star_col]).union(self.all_used_col)))
            # Resolve projections that have many columns, some of which could be *
            for p in projection.find_all(exp.Column):
                # Resolve * with other columns
                if isinstance(p, exp.Column) and p.find(exp.Star):
                    t_name = p.find(exp.Identifier).text("this")
                    # Resolve alias
                    if t_name in self.table_alias_dict.keys():
                        t_name = self.table_alias_dict[t_name]
                    # If from input table, get all columns from there
                    if t_name in self.input_table_dict.keys():
                        star_cols = self.input_table_dict[t_name]
                        # every column from there will be a column with that name
                        for per_star_col in star_cols:
                            target_dict[per_star_col] = sorted(
                                list(set(self._find_alias_col(col_sql=per_star_col, temp_table=source_table)).union(
                                    self.all_used_col)))
                    # If from another CTE, get all columns from there
                    elif t_name in self.cte_dict.keys():
                        star_cols = list(self.cte_dict[t_name].keys())
                        for per_star_col in star_cols:
                            target_dict[per_star_col] = sorted(
                                list(set(self.cte_dict[t_name][per_star_col]).union(self.all_used_col)))
                    # If from an unknown table, leave it with a STAR as temporary name
                    else:
                        target_dict[p.sql()] = [p.sql()] + (list(self.all_used_col))
                else:
                    # one projection can have many columns, append first
                    proj_columns.extend(self._find_alias_col(col_sql=p.sql(), temp_table=source_table))
            if proj_columns:
                target_dict[col_name] = sorted(list(set(proj_columns).union(self.all_used_col)))
        return target_dict

    def _resolve_table(self, part_ast: expressions = None) -> List:
        """
        Find the tables in the given ast
        :param part_ast: the ast to find the table
        """
        temp_table_list = []
        # Resolve FROM
        for table_sql in part_ast.find_all(exp.From):
            for table in table_sql.find_all(exp.Table):
                temp_table_list = self._find_table(table=table, temp_table_list=temp_table_list)
        # Resolve JOIN
        for table_sql in part_ast.find_all(exp.Join):
            for table in table_sql.find_all(exp.Table):
                temp_table_list = self._find_table(table=table, temp_table_list=temp_table_list)
        return temp_table_list

    def _find_table(self, table: expressions = None, temp_table_list: Optional[List] = None) -> List:
        """
        Update table alias and find all aliased used table names
        :param table: the expression with the table
        :param temp_table_list: temporary list of tables for appending the used tables
        :return:
        """
        if table.alias == "":
            self.table_alias_dict[table.sql()] = table.sql()
            temp_table_list.append(table.sql())
        else:
            temp = table.sql().split(" ")
            if temp[1] == "AS" or temp[1] == "as":
                self.table_alias_dict[temp[2]] = temp[0]
                temp_table_list.append(temp[0])
        return temp_table_list

    def _find_all_tables(self, temp_table_list: Optional[List] = None) -> List:
        """
        Update the used table names, such as if a CTE, update it with the dependant tables
        :param temp_table_list: temporary list of tables for appending the used tables
        :return:
        """
        ret_table = []
        for i in temp_table_list:
            table_name = i
            if i in self.table_alias_dict.keys():
                table_name = self.table_alias_dict[i]
            if table_name in self.cte_table_dict.keys():
                ret_table.extend(self.cte_table_dict[table_name])
            else:
                ret_table.append(table_name)
        return ret_table

    def _shared_col_conds(self, part_ast: expressions = None, used_tables: Optional[List] = None):
        """
        Extract all the columns in the shared conditions(WHERE, GROUP BY, etc)
        :param part_ast: the ast of the sql to extract
        :param used_tables: the tables that this sql uses
        """
        # COMBINE THE CONDITIONS
        for cond in shared_conditions:
            for cond_sql in part_ast.find_all(cond):
                for cond_col in cond_sql.find_all(exp.Column):
                    self.all_used_col.extend(self._find_alias_col(col_sql=cond_col.sql(), temp_table=used_tables))

    def _find_alias_col(self, col_sql: Optional[str] = "", temp_table: Optional[List] = None) -> List:
        """
        Find the columns and its alias and dependencies if it is from a cte
        :param col_sql: the sql to the column
        :param temp_table: the table that the sql uses
        :return:
        """
        temp = col_sql.split(".")
        # trying to deduce the table if all possible tables are eliminated
        elim_table = []
        if len(temp) < 2:
            for t in temp_table:
                if t in self.input_table_dict.keys():
                    if col_sql in self.input_table_dict[t]:
                        return [t + "." + col_sql]
                    else:
                        elim_table.append(t)
                elif t in self.cte_dict.keys():
                    if col_sql in self.cte_dict[t].keys():
                        return self.cte_dict[t][col_sql]
                    else:
                        elim_table.append(t)
            deduced_table = set(temp_table) - set(elim_table)
            if len(deduced_table) == 1:
                return [deduced_table.pop() + "." + col_sql]
        elif len(temp) == 2:
            if temp[0] in self.table_alias_dict.keys():
                t = self.table_alias_dict[temp[0]]
            else:
                t = temp[0]
            if t in self.cte_dict.keys():
                return self.cte_dict[t][temp[1]]
            else:
                return [t + "." + temp[1]]
        return [col_sql]

    def _resolve_agg_star(self, col_name: Optional[str] = "", projection: expressions = None, used_tables: Optional[List] = None):
        """
        Trying to resolve the * and append appropriate columns if the table is able to resolved
        :param col_name: the name of the column
        :param projection: the expression of the sql
        :param used_tables: the tables that are used
        """
        if projection.find(exp.Star):
            # * with a table name
            if projection.find(exp.Identifier):
                t_name = projection.find(exp.Identifier).text("this")
                # Resolve alias
                if t_name in self.table_alias_dict.keys():
                    t_name = self.table_alias_dict[t_name]
                if t_name in self.input_table_dict.keys():
                    star_cols = []
                    for s in self.input_table_dict[t_name]:
                        star_cols.extend(self._find_alias_col(col_sql=s, temp_table=used_tables))
                elif t_name in self.cte_dict.keys():
                    star_cols = []
                    for s in list(self.cte_dict[t_name].keys()):
                        star_cols.extend(self._find_alias_col(col_sql=s, temp_table=used_tables))
                else:
                    star_cols = [t_name + ".*"]
                self.column_dict[col_name] = sorted(list(set(star_cols).union(self.all_used_col)))
            # only star, like count(*)
            else:
                self.column_dict[col_name] = sorted(list(self.all_used_col))


if __name__ == "__main__":
    pass



from lineagex.utils import produce_json
from lineagex.SqlToDict import SqlToDict
from lineagex.ColumnLineageNoConn import ColumnLineageNoConn
from typing import Optional


class LineageXNoConn:
    def __init__(self, path: Optional[str] = "", search_schema: Optional[str] = "public") -> None:
        self.output_dict = {}
        search_schema = [x.strip() for x in search_schema.split(",")]
        self.sql_files_dict = SqlToDict(path, search_schema).sql_files_dict
        self.input_table_dict = {}
        self._run_lineage_no_conn()

    def _run_lineage_no_conn(self):
        """
        The driver function to extract the table lineage information
        :return: output an interactive html for the table lineage information
        """
        for name, sql in self.sql_files_dict.items():
            try:
                col_lineage = ColumnLineageNoConn(
                    sql=sql, input_table_dict=self.input_table_dict
                )
                self.output_dict[name] = {
                    "tables": col_lineage.table_list,
                    "columns": col_lineage.column_dict,
                    "table_name": name,
                }
                # add to the dict with the already parsed tables
                self.input_table_dict[name] = list(col_lineage.column_dict.keys())
            except Exception as e:
                print("{} is not processed because it countered {}".format(name, e))
                continue
        self._guess_schema_name()
        produce_json(self.output_dict)

    def _guess_schema_name(self):
        """
        Try to guess the schema names for the sql provided for a more accurate depiction
        """
        all_tables = []
        for key, val in self.output_dict.items():
            all_tables.extend(val['tables'])
        all_tables = list(set(all_tables))
        tables_dict = {}
        for t in all_tables:
            tables_dict[t.split(".")[-1]] = t
        for key, val in self.output_dict.copy().items():
            if key in tables_dict.keys():
                if tables_dict[key] != key:
                    self.output_dict[tables_dict[key]] = val
                    self.output_dict[tables_dict[key]]["table_name"] = tables_dict[key]
                    self.output_dict.pop(key)


if __name__ == "__main__":
    pass
